-- Find font to fit text tightly to a rectangle
--
-- Copyright © 2021 Pedro Gimeno Fortea
--
-- You can do whatever you want with this software, under the sole condition
-- that this notice and any copyright notices are preserved. It is offered
-- with no warrany, not even implied.

--[[

    Usage:
        local newFitter = require("tightFitText")

    Per font:
        local myFontFitter = newFitter("myFont.ttf")

    For the default Love2D font:
        local defaultFitter = newFitter()

    To use:
        myFontFitter:setFittingFont(text, width, height)
    will set the current font to one with a size such that the text fits
    tightly in a rectangle of the given width and height.

--]]


-- Table of precalculated size-to-pixels ratio.
-- Precalculate with a program which just does this:
-- print(love.graphics.newFont("name", 1000):getHeight()/1000)
local sizeRatios = {
  [false] = 1.164;
  ["LiberationMono-Regular.ttf"] = 1.133;
  ["LiberationSans-Regular.ttf"] = 1.150;
}

local newFont = love.graphics.newFont
local newFonts = 0

local function getFontAndHeight(self, size, name)
  local sizeName = size .. (name and "\0" .. name or "")
  if self.cacheFont[sizeName] == nil then
    if name then
      self.cacheFont[sizeName] = newFont(name, size)
    else
      self.cacheFont[sizeName] = newFont(size)
    end
    self.cacheHeight[sizeName] = self.cacheFont[sizeName]:getHeight()
  end
  return self.cacheFont[sizeName], self.cacheHeight[sizeName]
end

local function setFittingFont(self, text, width, height)
  local widthRef = self.fontRef:getWidth(text)
  if widthRef == 0 then return end
  local estimatedSize = math.ceil(math.min(self.sizeRef / widthRef * width,
    self.sizeRef / self.heightRef * height))
  local font, bestHeight = self:getFontAndHeight(estimatedSize, self.fontName)
  local bestWidth = font:getWidth(text)
  if bestWidth < width and bestHeight < height then
    -- Might still be too short; increase until found
    local oldFont
    repeat
      oldFont = font
      estimatedSize = estimatedSize + 1
      font, bestHeight = self:getFontAndHeight(estimatedSize, self.fontName)
      bestWidth = font:getWidth(text)
    until bestWidth > width or bestHeight > height
    estimatedSize = estimatedSize - 1
    font = oldFont
  elseif bestWidth > width or bestHeight > height then
    -- Decrease until found
    repeat
      estimatedSize = estimatedSize - 1
      font, bestHeight = self:getFontAndHeight(estimatedSize, self.fontName)
      bestWidth = font:getWidth(text)
    until bestWidth <= width and bestHeight <= height
  end
  love.graphics.setFont(font)
end

local function objNew(fontName, sizeRef)
  local self = {
    fontName = fontName;
    ratio = sizeRatios[fontName or false];
    cacheFont = {};
    cacheHeight = {};
    sizeRef = sizeRef or 100;
    setFittingFont = setFittingFont;
    getFontAndHeight = getFontAndHeight;
  }
  self.heightRef = self.sizeRef * self.ratio
  self.fontRef = self:getFontAndHeight(self.sizeRef, fontName)
  return self
end

return objNew
